% Version 1.000
%
% Code provided by Ruslan Salakhutdinov and Geoff Hinton
%
% Permission is granted for anyone to copy, use, modify, or distribute this
% program and accompanying programs and documents for any purpose, provided
% this copyright notice is retained and prominently displayed, along with
% a note saying that the original programs are available from our
% web page.
% The programs and documents are distributed without any warranty, express or
% implied.  As the programs were written for research purposes only, they have
% not been tested to the degree that would be advisable in any important
% application.  All use of these programs is entirely at the user's own risk.

% Version 1.100
%
% Updated by Computational Cognitive Neuroscience Lab
% University of Padova
% ccnl.psy.unipd.it
%
% Implementation on graphic processors (GPUs) using MATLAB Parallel Computing Toolbox

clear all; close all; clc

% DEEP NETWORK SETUP
% (parameters and final network weights will be saved in structure DN)
DN.layersize   = [100 100 500];        % network architecture
DN.nlayers     = length(DN.layersize);
DN.maxepochs   = 30;                    % unsupervised learning epochs
DN.batchsize   = 125;                   % mini-batch size
sparsity       = 1;                     % set to 1 to encourage sparsity on third layer
spars_factor   = 0.05;                  % how much sparsity?
epsilonw       = 0.1;                   % learning rate (weights)
epsilonvb      = 0.1;                   % learning rate (visible biases)
epsilonhb      = 0.1;                   % learning rate (hidden biases)
weightcost     = 0.0002;                % decay factor
init_momentum  = 0.5;                   % initial momentum coefficient
final_momentum = 0.9;                   % momentum coefficient

% load training dataset
fname = ['MNIST_data_' sprintf('%d',DN.batchsize) '.mat'];
load(fname);
fprintf(1,'\nUnsupervised training of a deep belief net\n');
DN.err = zeros(DN.maxepochs, DN.nlayers, 'single');
%fflush(stdout); % if running in Octave
tic();

for layer = 1:DN.nlayers
    
    % for the first layer, input data are raw images
    % for next layers, input data are preceding hidden activations
    fprintf(1,'Training layer %d...\n', layer);
    %fflush(stdout); % if running in Octave
    if layer == 1
        data = batchdata;
    else
        data  = batchposhidprobs;
    end
    
    % initialize weights and biases
    numhid  = DN.layersize(layer);
    [numcases, numdims, numbatches] = size(data);
    vishid       = 0.1*randn(numdims, numhid);
    hidbiases    = zeros(1,numhid);
    visbiases    = zeros(1,numdims);
    vishidinc    = zeros(numdims, numhid);
    hidbiasinc   = zeros(1,numhid);
    visbiasinc   = zeros(1,numdims);
    batchposhidprobs = zeros(DN.batchsize, numhid, numbatches);
    
    for epoch = 1:DN.maxepochs
        errsum = 0;
        for mb = 1:numbatches
            data_mb = data(:, :, mb);
            % learn an RBM with 1-step contrastive divergence
            rbm_CPU;
            errsum = errsum + err;
            if epoch == DN.maxepochs
                batchposhidprobs(:, :, mb) = poshidprobs;
            end
            if sparsity && (layer == 3)
                poshidact = sum(poshidprobs);
                Q = poshidact/DN.batchsize;
                if mean(Q) > spars_factor
                    hidbiases = hidbiases - epsilonhb*(Q-spars_factor);
                end
            end
        end
        DN.err(epoch, layer) = errsum;
    end
    % save learned weights
    DN.L{layer}.hidbiases  = hidbiases;
    DN.L{layer}.vishid     = vishid;
    DN.L{layer}.visbiases  = visbiases;
    
end

DN.learningtime = toc();
fprintf(1, '\nElapsed time: %d \n', DN.learningtime);
%fflush(stdout); % if running in Octave
fname = ['DBN_' sprintf('%d',DN.batchsize) '.mat'];
% save final network and parameters
save(fname, 'DN');
% if running Octave:
%save('-mat-binary', fname, 'DN');
clear all;